\\ ###########################################################
\\ rJPA trial programs for cubic field
\\ Programing language: PARI/GP
\\ Author     : Kiyohisa Tanimoto
\\ Issue date :2018/01/04
\\ License    : public domain
\\ Algorithms of rJPA:
\\   developped by Kiyohisa Tanimoto 
\\ Contained programs                        
\\ --------------------------------------------
\\  1. cubicrjpa                            
\\     one cycle of rJPA                    
\\  2. rjpacheck                            
\\     specified steps and peridicity check of rJPA 
\\  3. looprjpacheck                         
\\     rjpacheck within specified range      
\\  4. looprjpacheckr
\\     recursive version of rjpacheck 
\\  5. rjpaunit                              
\\     calculation of unit by rJPA           
\\  6. others                                
\\     components of the above programs      
\\
\\ Field definition polynomial limitaion     
\\  f(x) : monic polynomial of Z[x}  
\\  and f(x)=x^3+bx+c                        
\\  Usages                                   
\\ -------------------------------------------------
\\ 1. rjpacheck(p1,p2,p3);                
\\    p1 is the b coefficient of f(x)       
\\    p2 is the c coefficient of f(x)       
\\   p3 is the rjpa step specified length  
\\   --- output ------                      
\\   ---reach to periodic sequence case---  
\\  "max invrdet=" #max of 1/abs(rdet(m))   
\\  "periodic: period="  #period            
\\  "a unit is " # a derived unit           
\\   --- not yet reach periodic sequence case 
\\  "max invrdet=" #max of 1/abs(rdet(m))   
\\  "not yet reach to periodic sequence"    
\\  #last terms of rJPA tuples              
\\   ---f(x) is reducible case---           
\\   reducible                              
\\ 2. looprjpacheck(p1,p2,p3,p4);       
\\    p1 is the b coefficient of f(x)       
\\    p2 and p3 determine check range such as 
\\       c of f(X) is in {-p2,-p2-1,...,-p3} 
\\    p4 is the rJPA step specified length    
\\
\\   --- output ------                        
\\ loop head :following outputs are repeated  
\\  "c=" #c of f(x)                         
\\   ---reach to periodic sequence case---    
\\  "periodic: period="  #period              
\\  "a unit is " # a derived unit             
\\   --- not yet reach periodic sequence case 
\\  "not yet reach to periodic sequence"      
\\  #last terms of rJPA tuples               
\\   ---f(x) is reducible case---            
\\   reducible                               
\\ loop tail : specified times reapeated     
\\ --------------------------------------------------------
\\ 3.looprjpacheckr(p1,p2,p3,p4,p5)
\\     p1= b of definition polynomial x^3+bx+c=0
\\     p2= head of -c of definition polynomial x^3+bx+c=0
\\     p3= tail of -c of definition polynomial x^3+bx+c=0
\\     p4= the first length of rjpa sequence
\\     p5= the last length of rjpa sequence of p4*2^(p5-1)
\\  output image
\\    "---- try ----" loopcount "count=" #the number of remained objects "size=" #length of rjpa sequence
\\
\\ -------------------------------------------------------
\\ 4. rjpaunit(p1,p2)
\\      p1=b of definition polynomial x^3+bx+c=0
\\      p2=c of definition polynomial x^3+bx+c=0
\\   output image
\\    "max invrdet=" #the maximum of |1/rdet|
\\    "period="  #length of period
\\    "a unit is " a unit derived from rjpa
\\ ----------------------------------------------------------
\\ ##############################################################

cubicdisc(b,c)=-4*b^3 -27*c^2
cubicroot(b,c)=real(polroots(x^3+b*x+c)[1])
setvalue(bb,cc)={b1=bb; c1=cc; x1=cubicroot(bb,cc);return(x1)}
x1=cubicroot(-5,3)  \\ initial value for temporary use

roundeven(x)={if(x-floor(x)==0.5,return(2*round(x/2)),return(round(x)))}
\\ Bankers round function definition because PARI/GP doen't support it.

cubicvalue(p5)={local(a5,b5,c5);a5=p5[1];b5=p5[2];c5=p5[3];return(c5+x1*(b5+x1*a5))}

cubicround(p5)=roundeven(cubicvalue(p5))

cubicrdet(r5)={local(p5,q5,a5,b5,a6,b6);p5=r5[1];q5=r5[2];a5=p5[1];b5=p5[2];a6=q5[1];b6=q5[2];return(a5*b6-a6*b5)};

cubicinvrdet(r5)=abs(1/cubicrdet(r5))

cubicinv(p5)={local(a5,b5,c5,a7,b7,c7,d7);a5=p5[1];b5=p5[2];c5=p5[3];d7=-a5^2*b5*b1*c1-b5^3*c1+a5^3*c1^2+a5^2*c5*b1^2+b5^2*c5*b1+3*a5*b5*c5*c1-2*a5*c5^2*b1+c5^3; a7=(a5^2*b1-a5*c5+b5^2)/d7; b7=(-a5^2*c1-b5*c5)/d7; c7=(c5^2+a5^2*b1^2-2*a5*c5*b1+a5*b5*c1+b5^2*b1)/d7; return([a7,b7,c7])}

cubicmul(p5,q5)={local(a5,b5,c5,a6,b6,c6,a9,b9,c9);a5=p5[1];b5=p5[2];c5=p5[3];a6=q5[1];b6=q5[2];c6=q5[3];a9=a5*c6+a6*c5+b5*b6-a5*a6*b1; b9=b5*c6+b6*c5-(a5*b6+a6*b5)*b1-a5*a6*c1; c9=c5*c6-(a5*b6+a6*b5)*c1; return([a9,b9,c9])}

cubicdiv(p5,q5)=cubicmul(p5,cubicinv(q5))

cubicsc1(p5)={local(a5,b5,c5,u5);a5=p5[1];b5=p5[2];c5=p5[3];u5=cubicround(p5);return([a5,b5,c5-u5])}
\\ a part of SC-step of rJPA

cubicrc(r5)={local(p5,q5,a5,b5,c5,a6,b6,c6,a7,b7,c7,w1,w2); p5=r5[1];q5=r5[2];a5=p5[1];b5=p5[2];c5=p5[3];a6=q5[1];b6=q5[2];c6=q5[3];
until(w1==0 && w2==0,w1=roundeven((a5*a6+b5*b6)/(a5^2+b5^2));a6=a6-w1*a5;b6=b6-w1*b5;c6=c6-w1*c5;w2=roundeven((a5*a6+b5*b6)/(a6^2+b6^2));a5=a5-w2*a6;b5=b5-w2*b6;c5=c5-w2*c6);
if(a5^2+b5^2>a6^2+b6^2,return([[a6,b6,c6],[a5,b5,c5]]),return([[a5,b5,c5],[a6,b6,c6]]))}
\\ RC-step of rJPA

basepair=[[0,1,0],[1,0,0]]

cubicrjpa(r5)={local(r6,k1,k2,k3); 
  r6=cubicrc(r5);
  k1=cubicsc1(r6[1]);
  if(cubicvalue(k1)<0,k1=-k1);  \\ normalization
  k3=cubicsc1(r6[2]);
  if(cubicvalue(k3)<0,k3=-k3);  \\ 
  k1=cubicinv(k1);
  return([cubicmul(k3,k1),k1]) }
\\ usage: cubicrjpa([[p1,p2,p3],[q1,q2,q3]])

nestlist(func,obj,nn)={local(templist,i,tempobj);
  templist=[]; tempobj=obj;
  for(i=1,nn,
    templist=concat(tempobj,templist);
    tempobj=func(tempobj)  );
  return(templist)  }
  

\\ usage rjpacheck(bb,cc,nn)
rjpacheck(bb,cc,nn)={local(jj,pp0,qq0,yjpalist,lastset,flag,invrdet,len,testp,jpaunit);
  if(cubicdisc(bb,cc)==0,print("reducible"),setvalue(bb,cc);
     invrdet=1;
     flag=0;
     testp=x^3+bb*x+cc;
     if(polisirreducible(testp)==0,print("reducible"),
       yjpalist=[];
       jpaunit=[0,0,1];
       lastset=basepair;
       for(jj=1,nn,
         yjpalist=concat(lastset,yjpalist);
         lastset=cubicrjpa(lastset);
         if(invrdet<cubicinvrdet(lastset),invrdet=cubicinvrdet(lastset));
          );
       print("max invrdet=",invrdet);
       pp0=lastset[1];  qq0=lastset[2]; len=0;
       for(jj=1,nn,
         if(pp0==yjpalist[2*jj-1],if(qq0==yjpalist[2*jj],flag=1));
         jpaunit=cubicmul(jpaunit,yjpalist[2*jj]);
         if(flag==1,break,len=len+1);
         );
       if(flag==1,print("periodic: period=",len+1);print("a unit is ",cubicinv(jpaunit)),print("not yet reach to periodic sequence ",lastset) );
      );
    );   }

\\ usage: looprjpacheck(b,start of -c, end of -c, sequence length)
looprjpacheck(bb,ccs,cce,nn)={local(cc,jj,pp0,qq0,yjpalist,lastset,flag,invrdet,len,testp);
  forstep(cc=-ccs,-cce,-1,
  if(cubicdisc(bb,cc)==0,print("reducible"),setvalue(bb,cc);
     invrdet=1;
     flag=0;
     testp=x^3+bb*x+cc;
     print("c=", cc);
     if(polisirreducible(testp)==0,print("reducible"),
       yjpalist=[];
       lastset=basepair;
       for(jj=1,nn,
         yjpalist=concat(lastset,yjpalist);
         lastset=cubicrjpa(lastset);
         if(invrdet<cubicinvrdet(lastset),invrdet=cubicinvrdet(lastset));
          );
       print("max invrdet=",invrdet);
       pp0=lastset[1];  qq0=lastset[2]; len=0;
       for(jj=1,nn,
         if(pp0==yjpalist[2*jj-1],if(qq0==yjpalist[2*jj],flag=1));
         if(flag==1,break,len=len+1);
         );
       if(flag==1,print("periodic: period=",len+1),print("not yet periodic");print(lastset));
      );  \\ polisireeducible end
    );  );  }    \\if cubiddisc end  for step end


looprjpacheck2(bb,ccs,cce,nn)={local(cc,jj,pp0,qq0,yjpalist,lastset,flag,invrdet,len,testp);
  forstep(cc=-ccs,-cce,-1,
  if(cubicdisc(bb,cc)==0,print("reducible"),setvalue(bb,cc);
     invrdet=1;
     flag=0;
     testp=x^3+bb*x+cc;
     if(cc % 100 ==0,print(cc,"  case") );
     if(polisirreducible(testp)==1,
       yjpalist=[];
       lastset=basepair;
       for(jj=1,nn,
         yjpalist=concat(lastset,yjpalist);
         lastset=cubicrjpa(lastset);
         if(invrdet<cubicinvrdet(lastset),invrdet=cubicinvrdet(lastset));
          );
       pp0=lastset[1];  qq0=lastset[2]; len=0;
       for(jj=1,nn,
         if(pp0==yjpalist[2*jj-1],if(qq0==yjpalist[2*jj],flag=1));
         if(flag==1,break,len=len+1);
         );
       if(flag==0,print(cc, " case :not yet periodic, max invrdet=",invrdet);print(lastset));
      ); \\ if polisireeducible end
    );  );  }


looprjpacheckr(bb,ccs,cce,nn,depth)={local(cc,jj,len,pp0,qq0,yjpalist,retrylist,nretrylist,lastset,flag,rlen,testp,retryp,dcnt,retrysize);
  retrylist=[]; retrysize=nn;
  for(cc=-cce,-ccs,
   testp=x^3+bb*x+cc;
   if(polisirreducible(testp)==1,retrylist=concat([cc],retrylist) )
  );
 for(dcnt=1,depth,
  nretrylist=[];
  rlen=length(retrylist);
  print("------------ try ----", dcnt," count=",rlen,"  length=",retrysize);
  if(rlen==0,,
   for(retryp=1,rlen,
     cc=retrylist[retryp];
     setvalue(bb,cc);
     flag=0;
\\     if(cc % 100 ==0,print(cc,"  case") );
     yjpalist=[];
     lastset=basepair;
     for(jj=1,retrysize,
       yjpalist=concat(lastset,yjpalist);
       lastset=cubicrjpa(lastset);
     );
     pp0=lastset[1];  qq0=lastset[2]; len=0;
     for(jj=1,retrysize,
         if(pp0==yjpalist[2*jj-1],if(qq0==yjpalist[2*jj],flag=1));
         if(flag==1,break,len=len+1);
      );
      if(flag==1,,nretrylist=concat([cc],nretrylist));
\\     if(flag==1,print(cc,"  periodic :period=",len),print(cc, " case :not yet periodic  ");print(lastset);nretrylist=concat([cc],nretrylist)); 
     );   \\ for rlen loop
    );  \\ if rlen loop
    retrylist=nretrylist; retrysize=2*retrysize;
    );   \\ dcnt loop
  }

\\ usage rjpaunit(bb,cc)
rjpaunit(bb,cc)={local(jj,pp0,qq0,yjpalist,lastset,flag,invrdet,len,testp,jpaunit,nn,dcnt,depth,listlen);
  if(cubicdisc(bb,cc)==0,print("reducible"),setvalue(bb,cc);
     nn=10;
     listlen=0;
     invrdet=1;
     flag=0;
     depth=16;
     testp=x^3+bb*x+cc;
     if(polisirreducible(testp)==0,print("reducible"),
       yjpalist=[];
       lastset=basepair;
    for(dcnt=1,depth,
       jpaunit=[0,0,1];
       for(jj=1,nn,
         yjpalist=concat(lastset,yjpalist);
         lastset=cubicrjpa(lastset);
         if(invrdet<cubicinvrdet(lastset),invrdet=cubicinvrdet(lastset));
          );    \\ 1st for loop end
       pp0=lastset[1];  qq0=lastset[2]; len=0;
      listlen=listlen+nn;
       for(jj=1,listlen,
         if(pp0==yjpalist[2*jj-1],if(qq0==yjpalist[2*jj],flag=1));
         jpaunit=cubicmul(jpaunit,yjpalist[2*jj]);
         if(flag==1,break,len=len+1);
         );
       if(flag==1,break,nn=listlen);
    );  print("max invrdet=",invrdet);print("period=",len+1);print("unit:",cubicinv(jpaunit));  \\ dcnt loop end
    );  ); }

help()={
print("rJPA test programs for PARI/GP   Ver. 1.0");
print("    written by Kiyohisa Tanimoto");
print("     Ver. 1.0 on 2018/01/04");
print(" license type: public domain    ");
print("---------------------------------------");
print(" the following functios are available");
print(" 1.cubicrjpa([[p1,p2,p3],[q1,q2,q3]])  rJPA one cycle");
print(" 2.rjpacheck(p1,p2,p3)                 rjpa algorithm");
print(" 3.looprjpacheck(p1,p2,p3,p4)          series of test data for rjpa"); 
print(" 4.looprjpacheckr(p1,p2,p3,p4,p5)      recursive type of looprjpacheck");
print(" 5.rjpaunit(p1,p2)                     a unit calculation by rjpa ");
}

help()


\\  end of programs

